const x_PI = 3.14159265358979324 * 3000.0 / 180.0
const PI = 3.1415926535897932384626
const a = 6378245.0 // 地球长半轴
const ee = 0.00669342162296594323


function bd09togcj02(lng, lat) {
    const bd_lon = Number(lng);
    const bd_lat = Number(lat);
    const x_pi = 3.14159265358979324 * 3000.0 / 180.0;
    const x = bd_lon - 0.0065;
    const y = bd_lat - 0.006;
    const z = Math.sqrt(x * x + y * y) - 0.00002 * Math.sin(y * x_pi);
    const theta = Math.atan2(y, x) - 0.000003 * Math.cos(x * x_pi);
    const gg_lng = z * Math.cos(theta);
    const gg_lat = z * Math.sin(theta);
    return {
        lng: gg_lng,
        lat: gg_lat
    }
}

function bd09towgs84(bd_lon, bd_lat) {
    const res = bd09togcj02(bd_lon, bd_lat);
    return gcj02towgs84(res.lng, res.lat)
}

function wgs84tobd09(lng, lat) {
    const res = wgs84togcj02(lng, lat);
    return gcj02tobd09(res.lng, res.lat)
}

function gcj02tobd09(lng, lat) {
    lng = Number(lng)
    lat = Number(lat)
    const z = Math.sqrt(lng * lng + lat * lat) + 0.00002 * Math.sin(lat * x_PI);
    const theta = Math.atan2(lat, lng) + 0.000003 * Math.cos(lng * x_PI);
    const bd_lng = z * Math.cos(theta) + 0.0065;
    const bd_lat = z * Math.sin(theta) + 0.006;
    return {lng: Number(bd_lng), lat: Number(bd_lat)}
}

function wgs84togcj02(lng, lat) {
    lng = Number(lng)
    lat = Number(lat)
    let dlat = transformlat(lng - 105.0, lat - 35.0);
    let dlng = transformlng(lng - 105.0, lat - 35.0);
    const radlat = lat / 180.0 * PI;
    let magic = Math.sin(radlat);
    magic = 1 - ee * magic * magic
    const sqrtmagic = Math.sqrt(magic);
    dlat = (dlat * 180.0) / ((a * (1 - ee)) / (magic * sqrtmagic) * PI)
    dlng = (dlng * 180.0) / (a / sqrtmagic * Math.cos(radlat) * PI)
    const mglat = lat + dlat;
    const mglng = lng + dlng;
    return {
        lng: mglng,
        lat: mglat,
    }
}

function gcj02towgs84(lng, lat) {
    lng = Number(lng)
    lat = Number(lat)
    let dlat = transformlat(lng - 105.0, lat - 35.0);
    let dlng = transformlng(lng - 105.0, lat - 35.0);
    const radlat = lat / 180.0 * PI;
    let magic = Math.sin(radlat);
    magic = 1 - ee * magic * magic
    const sqrtmagic = Math.sqrt(magic);
    dlat = (dlat * 180.0) / ((a * (1 - ee)) / (magic * sqrtmagic) * PI)
    dlng = (dlng * 180.0) / (a / sqrtmagic * Math.cos(radlat) * PI)
    const mglat = lat + dlat;
    const mglng = lng + dlng;
    return {
        lng: lng * 2 - mglng,
        lat: lat * 2 - mglat,
    }
}

function transformlat(lng, lat) {
    lng = Number(lng)
    lat = Number(lat)
    let ret = -100.0 + 2.0 * lng + 3.0 * lat + 0.2 * lat * lat + 0.1 * lng * lat + 0.2 * Math.sqrt(Math.abs(lng));
    ret += (20.0 * Math.sin(6.0 * lng * PI) + 20.0 * Math.sin(2.0 * lng * PI)) * 2.0 / 3.0
    ret += (20.0 * Math.sin(lat * PI) + 40.0 * Math.sin(lat / 3.0 * PI)) * 2.0 / 3.0
    ret += (160.0 * Math.sin(lat / 12.0 * PI) + 320 * Math.sin(lat * PI / 30.0)) * 2.0 / 3.0
    return ret
}

function transformlng(lng, lat) {
    lng = Number(lng)
    lat = Number(lat)
    let ret = 300.0 + lng + 2.0 * lat + 0.1 * lng * lng + 0.1 * lng * lat + 0.1 * Math.sqrt(Math.abs(lng));
    ret += (20.0 * Math.sin(6.0 * lng * PI) + 20.0 * Math.sin(2.0 * lng * PI)) * 2.0 / 3.0
    ret += (20.0 * Math.sin(lng * PI) + 40.0 * Math.sin(lng / 3.0 * PI)) * 2.0 / 3.0
    ret += (150.0 * Math.sin(lng / 12.0 * PI) + 300.0 * Math.sin(lng / 30.0 * PI)) * 2.0 / 3.0
    return ret
}

function transformCoordinate(local, current, target, out = 'array') {
    const res = JSON.parse(JSON.stringify(local));
    if (!current) current = 'EPSG3857';
    if (!target) target = 'EPSG3857';
    current = current.replace(/EPSG:?/g, 'EPSG');
    target = target.replace(/EPSG:?/g, 'EPSG');
    if(current === 'CGCS2000') current = 'EPSG4326';
    if(target === 'CGCS2000') target = 'EPSG4326';
    const x = res[0] || res.x || res.lng || 0;
    const y = res[1] || res.y || res.lat || 0;
    if (current === target) {
        return out === 'array' ? [x, y] : {x, y};
    }
    let transformFn;
    if (current === 'EPSG3857') {
        if (target === 'EPSG4326') {
            transformFn = gcj02towgs84;
        } else if (target === 'baidu') {
            transformFn = gcj02tobd09;
        }
    } else if (current === 'EPSG4326') {
        if (target === 'EPSG3857') {
            transformFn = wgs84togcj02;
        } else if (target === 'baidu') {
            transformFn = wgs84tobd09;
        }
    } else if (current === 'baidu') {
        if (target === 'EPSG3857') {
            transformFn = bd09togcj02;
        } else if (target === 'EPSG4326') {
            transformFn = bd09towgs84;
        }
    }
    if (!transformFn) Error(current + 'to' + target + '未支持的坐标转换');
    const result = transformFn(x, y);
    return out === 'array' ? [result.lng, result.lat] : {x: result.lng, y: result.lat}
}

function transformCoordinates(coordinates = [], current, target, out = 'array') {
    const resultSet = [];
    let len = coordinates.length;
    for (let i = 0; i < len; i++) {
        resultSet.push(transformCoordinate(coordinates[i], current, target, out = 'array'))
    }
    return resultSet;
}

function getUuid() {
    const len = 32; //32长度
    let radix = 16; //16进制
    const chars = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz'.split('');
    const uuid = [];
    let i;
    radix = radix || chars.length;
    if (len) {
        for (i = 0; i < len; i++) {
            uuid[i] = chars[0 | Math.random() * radix];
        }
    } else {
        let r;
        uuid[8] = uuid[13] = uuid[18] = uuid[23] = '-';
        uuid[14] = '4';
        for (i = 0; i < 36; i++) {
            if (!uuid[i]) {
                r = 0 | Math.random() * 16;
                uuid[i] = chars[i === 19 ? r & 0x3 | 0x8 : r];
            }
        }
    }
    return uuid.join('');
}

function divideString(str = '', size) {
    const result = [];
    const num = Math.ceil(str.length / size);
    for (let i = 0; i < num; i++) {
        let _;
        if (i === 0) {
            _ = str.slice(0, (i + 1) * size + 1)
        } else {
            _ = str.slice(i * size + 1, (i + 1) * size + 1)
        }
        result.push(_)
    }
    return result
}

function copy(obj) {
    return JSON.parse(JSON.stringify(obj));
}
